print("import: 'msrest'")
import msrest

print("import: 'msrest.authentication'")
import msrest.authentication

print("import: 'msrest.exceptions'")
import msrest.exceptions

print("import: 'msrest.pipeline'")
import msrest.pipeline

